import glob
import json
import logging
import math
import os
import threading

import maya.api.OpenMaya as OM
import maya.cmds

import larian.physics.model
import larian.physics.serializer

import exporter_common as common


reload(larian.physics.model)  # noqa
reload(larian.physics.serializer)  # noqa

reload(common)  # noqa


_MAX_FLOAT_PRECISION = 4
_NODE_TYPES = ['bulletRigidBodyShape', 'bulletRigidBodyConstraintShape']


def run(layer_name, path, output_type):
    return _export(layer_name, path, output_type)


def _export(layer_name, path, output_type_string):

    scene = _make_scene(layer_name)

    if not scene:
        logging.error("Can't build scene from nodes")
        return False

    scene_json = larian.physics.serializer.save(scene)

    if not scene_json:
        logging.error("Can't serialize scene to json structure")
        return False

    data = json.dumps(scene_json)

    return common.write_export_data(data, path, output_type_string)


def _make_scene(layer_name):
    mel_paths = maya.cmds.editDisplayLayerMembers(layer_name, query=True, fullNames=True)

    selection_list = OM.MSelectionList()

    for path in mel_paths:
        selection_list.add(path)

    scene = larian.physics.model.Scene()

    scene.coordinate_system.x = larian.physics.model.EAxisDirection.RIGHT
    scene.coordinate_system.y = larian.physics.model.EAxisDirection.UP
    scene.coordinate_system.z = larian.physics.model.EAxisDirection.FORWARD

    success = True

    for i in range(selection_list.length()):
        dag_path = selection_list.getDagPath(i)
        success &= _recurse_into_node(scene, dag_path.node())

    return scene if success else None


def _recurse_into_node(scene, node):
    if not node.hasFn(OM.MFn.kDagNode):
        return True

    node = OM.MFnDagNode(node)
    name = _get_node_name(node)

    if node.typeName == 'bulletRigidBodyShape':
        body = _make_rigid_body(name, node)

        if body is None:
            return False

        scene.bodies[name] = body

        return True

    if node.typeName == 'bulletRigidBodyConstraintShape':
        constraint = _make_constraint(name, node)

        if constraint is None:
            return False

        scene.constraints[name] = constraint

        return True

    success = True

    childCount = node.childCount()

    for i in range(childCount):
        success &= _recurse_into_node(scene, node.child(i))

    return success


def _make_rigid_body(name, node):
    body = larian.physics.model.RigidBody()

    body.type = _get_enum(node, 'bodyType').split(' ')[0]
    body.mass = _get_float(node, "mass")
    body.friction = _get_float(node, "friction")
    body.transform = _make_transform(node)
    body.shape = _make_shape(name, node)

    if body.shape is None:
        return None

    return body


def _make_shape(name, node):
    shape_type = _get_enum(node, "colliderShapeType")

    try:
        maker = globals()['_make_shape_' + shape_type]
    except KeyError:
        logging.error("{}: Unsupported shape type: {}".format(name, shape_type))
        return None

    shape = maker(node)
    shape.type = shape_type

    return shape


def _make_shape_box(node):
    box = larian.physics.model.ShapeBox()
    box.extents = _make_linear_vector3(node, 'extents')
    return box


def _make_shape_sphere(node):
    sphere = larian.physics.model.ShapeSphere()
    sphere.radius = _make_coordinate(node, 'radius')
    return sphere


def _make_shape_capsule(node):
    radius = _make_coordinate(node, 'radius')

    capsule = larian.physics.model.ShapeCapsule()

    capsule.axis = _get_enum(node, "axis")

    capsule.radius = radius
    capsule.length = round(_make_coordinate(node, 'length') - radius, _MAX_FLOAT_PRECISION)

    return capsule


def _make_shape_cylinder(node):
    cylinder = larian.physics.model.ShapeCylinder()

    cylinder.axis = _get_enum(node, "axis")

    cylinder.radius = _make_coordinate(node, 'radius')
    cylinder.length = _make_coordinate(node, 'length')

    return cylinder


def _make_shape_hull(node):
    hull = larian.physics.model.ShapeHull()

    node = _get_connected_node(node, "inMesh")

    mesh = OM.MFnMesh(node.object())

    points = mesh.getPoints()

    for point in points:
        vertex = larian.physics.model.Vector3()
        vertex.x = _as_coordinate(point.x)
        vertex.y = _as_coordinate(point.y)
        vertex.z = _as_coordinate(point.z)
        hull.points.append(vertex)

    return hull


def _make_constraint(name, node):
    type = _get_enum(node, "constraintType")

    if type != "sixdof":
        logging.error("{}: Unsupported constraint type: {}".format(name, type))

    constraint = larian.physics.model.Constraint()

    constraint.body_a_name = _get_object_name(node, "rigidBodyA")
    constraint.body_b_name = _get_object_name(node, "rigidBodyB")

    constraint.transform = _make_transform(node)

    constraint.linear_limits, constraint.angular_limits = _make_constraint_limits(node)

    return constraint


def _make_transform(node):
    parent_path = node.getPath()
    parent_path.pop()

    assert parent_path.hasFn(OM.MFn.kTransform)

    parent_node = OM.MFnTransform(parent_path)

    basis = parent_node.rotation(OM.MSpace.kWorld, True).asMatrix()
    position = parent_node.translation(OM.MSpace.kWorld)

    transform = larian.physics.model.Transform()

    transform.translation.x = _as_coordinate(position.x)
    transform.translation.y = _as_coordinate(position.y)
    transform.translation.z = _as_coordinate(position.z)

    for row in (0, 1, 2):
        for column in (0, 1, 2):
            transform.basis[row][column] = basis.getElement(row, column)

    return transform


def _make_constraint_limits(node):
    angular_limits = larian.physics.model.Constraint.Limits()

    angular_limits.lower = _make_angular_vector3(node, "angularConstraintMin")
    angular_limits.upper = _make_angular_vector3(node, "angularConstraintMax")

    linear_limits = larian.physics.model.Constraint.Limits()

    linear_limits.lower = _make_linear_vector3(node, "linearConstraintMin")
    linear_limits.upper = _make_linear_vector3(node, "linearConstraintMax")

    angular_locks = _get_enum3(node, "angularConstraint")
    linear_locks = _get_enum3(node, "linearConstraint")

    _process_limit_locks(angular_limits, angular_locks)
    _process_limit_locks(linear_limits, linear_locks)

    return linear_limits, angular_limits


def _process_limit_locks(limits, locks):
    for axis, kind in zip(('x', 'y', 'z'), locks):
        if kind == "free":
            setattr(limits.upper, axis, -1.0)
            setattr(limits.lower, axis, 1.0)
        elif kind == "locked":
            setattr(limits.upper, axis, 0.0)
            setattr(limits.lower, axis, 0.0)


def _make_linear_vector3(node, name):
    coords = _get_coord_array(node, name, 3)

    vector3 = larian.physics.model.Vector3()

    vector3.x = _as_coordinate(coords[0])
    vector3.y = _as_coordinate(coords[1])
    vector3.z = _as_coordinate(coords[2])

    return vector3


def _make_angular_vector3(node, name):
    coords = _get_coord_array(node, name, 3)

    vector3 = larian.physics.model.Vector3()

    vector3.x = _as_angle(coords[0])
    vector3.y = _as_angle(coords[1])
    vector3.z = _as_angle(coords[2])

    return vector3


def _make_coordinate(node, name):
    return _as_coordinate(_get_float(node, name))


def _make_angle(node, name):
    return _as_angle(_get_float(node, name))


def _make_float(node, name):
    return round(_get_float(node, name), _MAX_FLOAT_PRECISION)


def _as_coordinate(scalar):
    # Maya's bullet is ALWAYS in cm
    return round(scalar * 0.01, _MAX_FLOAT_PRECISION)


def _as_angle(scalar):
    return round(scalar, _MAX_FLOAT_PRECISION)


def _as_angle_convert_to_degrees(scalar):
    return round(math.degrees(scalar), _MAX_FLOAT_PRECISION)


def _get_coord_array(node, name, count):
    plug = node.findPlug(name, False)

    assert plug.isCompound and plug.numChildren() == count

    return [plug.child(i).asFloat() for i in range(count)]


def _get_float(node, name):
    plug = node.findPlug(name, False)
    return plug.asFloat()


def _get_int(node, name):
    plug = node.findPlug(name, False)
    return plug.asInt()


def _get_node_name(node):
    path = node.getPath()
    path.pop()

    name = OM.MFnDagNode(path).name()

    if name.startswith('|'):
        name = name[1:]

    name = name.replace('|', '.')

    return name.encode('utf8')


def _get_connected_node(node, name):
    plug = node.findPlug(name, False)

    assert plug.isConnected and plug.isDestination and not plug.isSource

    connection = plug.connectedTo(plug.isDestination, plug.isSource)[0]
    node = OM.MFnDagNode(connection.node())

    return node


def _get_object_name(node, name):
    node = _get_connected_node(node, name)
    return _get_node_name(node)


def _get_enum(node, name):
    plug = node.findPlug(name, False)

    attribute = plug.attribute()
    enum = OM.MFnEnumAttribute(attribute)

    index = plug.asInt()

    return enum.fieldName(index).encode('utf8').lower()


def _get_enum3(node, name):
    # For some reason you can't get 'name' plug directly, like in float3 above
    return (
        _get_enum(node, name + 'X'),
        _get_enum(node, name + 'Y'),
        _get_enum(node, name + 'Z')
    )


def _run_stderr_logger(fd):
    def _loop():
        while True:
            line = fd.read()

            if line == '':
                return

            line = line.rstrip()
            logging.error(line)

    reader = threading.Thread(target=_loop)
    reader.daemon = True
    reader.start()

    return reader


def _find_python3_interpreter():
    for prefix in os.environ['PATH'].split(os.pathsep):
        if len(glob.glob(os.path.join(prefix, 'python3*.dll'))) > 0:
            return os.path.join(prefix, 'pythonw.exe')

    return None
