import math


# Math is always in left-hand rotation rule, right-hand chirality, yxz up-right-view map


class Vector(object):
    def __init__(self, *args):
        self.values = [0] * 3

        for i, coord in enumerate(args):
            self.values[i] = coord

    @property
    def x(self):
        return self.values[0]

    @x.setter
    def _set_x(self, value):
        self.values[0] = value

    @property
    def y(self):
        return self.values[1]

    @y.setter
    def _set_y(self, value):
        self.values[1] = value

    @property
    def z(self):
        return self.values[2]

    @z.setter
    def _set_z(self, value):
        self.values[2] = value

    def __getitem__(self, i):
        return self.values[i]

    def __setitem__(self, i, value):
        self.values[i] = value

    def __additem__(self, i, value):
        self.values[i] += value


class Matrix(object):
    def __init__(self, *args):
        self.values = [0] * 9

        for i, x in enumerate(args):
            self.values[i] = x

    def __getitem__(self, coords):
        row, column = coords
        return self.values[row*3 + column]

    def __setitem__(self, coords, value):
        row, column = coords
        self.values[row*3 + column] = value

    def __additem__(self, coords, value):
        row, column = coords
        self.values[row*3 + column] += value


def make_matrix_from_rotation(rotation, order):
    matrix = Matrix()

    left = _make_primitive_matrix(order, rotation, 0)
    center = _make_primitive_matrix(order, rotation, 1)
    right = _make_primitive_matrix(order, rotation, 2)

    # only single value at left.index, right.index intersection, and it's sin(center) with it's sign

    main_sin = center[left.index, right.index]
    matrix[left.index, right.index] = main_sin

    # simple row at left.index, use cos from center and sin/cos from right

    for i in (0, 1, 2):
        matrix[left.index, i] += math.cos(center.angle) * right[left.index, i]

    # simple column at right.index, use cos from center and sin/cos from left

    for i in (0, 1, 2):
        matrix[i, right.index] += math.cos(center.angle) * left[i, right.index]

    # Other cells are composed of
    #
    # <left (center-index) column> joined (multiplied) cell-by-cell with <right (center-index) row>
    #
    #                                   left_column                               right_column
    # up_row:   | left(up_row,   center-index) right(center-index, left_column) |     ...      |
    # down_row: | left(down_row, center-index) right(center-index, left_column) |     ...      |
    #
    # And added:
    #
    # <left (right-index) column> joined (multiplied) cell-by-cell with <right (left-index) row> and -main_sin
    #                                   left_column                                        right_column
    # up_row:   | + left(up_row,   right-index) right(left-index, left_column) -main_sin |     ...      |
    # down_row: | + left(down_row, right-index) right(left-index, left_column) -main_sin |     ...      |
    #

    down_row = min(right.index, center.index)
    up_row = max(right.index, center.index)

    left_column = min(left.index, center.index)
    right_column = max(left.index, center.index)

    for row in (down_row, up_row):
        for column in (left_column, right_column):
            matrix[row, column] = left[row, center.index] * right[center.index, column] \
                                + left[row, right.index] * right[left.index, column] * (-main_sin)

    return matrix


def make_rotation_from_matrix(matrix, order):
    # sin as at (center-index) (center-index)

    left = _get_primitive_spec(order, 0)
    center = _get_primitive_spec(order, 1)
    right = _get_primitive_spec(order, 2)

    # Center sin is the start point

    center_sin = matrix[left.index, right.index] * center.get_sin_sign(left.index, right.index)

    center_angle = math.asin(center_sin)

    # Now (left-index) row will have cos(center) and sin/cos(right) in (center-index), (left-index)

    right_sin = matrix[left.index, center.index] * right.get_sin_sign(left.index, center.index)
    right_cos = matrix[left.index, left.index]

    right_angle = math.atan2(right_sin, right_cos)

    # Similarly (right-index) column have sin/cos(left) in (right-index), (center-index) rows

    left_sin = matrix[center.index, right.index] * left.get_sin_sign(center.index, right.index)
    left_cos = matrix[right.index, right.index]

    left_angle = math.atan2(left_sin, left_cos)

    return _make_rotation(order, left_angle, center_angle, right_angle)


def apply_matrix_to_vector(matrix, vector):
    transformed_vector = Vector()

    for i in (0, 1, 2):
        for j in (0, 1, 2):
            transformed_vector[i] += matrix[i, j] * vector[j]

    return transformed_vector


_CELL_ZERO = 0
_CELL_ONE = 1
_CELL_COS = 2
_CELL_SIN = 3


# Generic transformation is
# cos(a) sin(a)
# -sin(a) cos(a)
#
# at axis row and column there are zeros, at intersection = 1
# sin sign is negative if (row > column), flipped for y (odd axes)
#
class _PrimitiveRotationSpec(object):
    def __init__(self, index, left_hand_sin_sign):
        self.index = index
        self.left_hand_sin_sign = left_hand_sin_sign

    def make_matrix(self, angle):
        return _PrimitiveRotationMatrix(self, angle)

    def get_cell_type(self, row, column):
        row_is_axis = row == self.index
        column_is_axis = column == self.index

        if row_is_axis and column_is_axis:
            return _CELL_ONE
        elif row_is_axis or column_is_axis:
            return _CELL_ZERO
        elif row == column:
            return _CELL_COS
        else:
            return _CELL_SIN

    def get_sin_sign(self, row, column):
        return self.left_hand_sin_sign * (+1 if row < column else -1)


class _PrimitiveRotationMatrix(object):
    def __init__(self, spec, angle):
        self.spec = spec
        self.angle = angle

    @property
    def index(self):
        return self.spec.index

    def __getitem__(self, coords):
        row, column = coords

        t = self.spec.get_cell_type(row, column)

        if t == _CELL_ONE:
            return 1
        elif t == _CELL_ZERO:
            return 0
        elif t == _CELL_COS:
            return math.cos(self.angle)
        else:
            assert t == _CELL_SIN

            sign = self.spec.get_sin_sign(row, column)
            return sign * math.sin(self.angle)


_PRIMITIVE_MATRIX_SPECS = [
    _PrimitiveRotationSpec(0, +1),
    _PrimitiveRotationSpec(1, -1),
    _PrimitiveRotationSpec(2, +1)
]


def _get_primitive_spec(order, index):
    axis_index = ord(order[index]) - ord('x')

    return _PRIMITIVE_MATRIX_SPECS[axis_index]


def _make_primitive_matrix(order, rotation, index):
    spec = _get_primitive_spec(order, index)

    angle = rotation[spec.index]

    return spec.make_matrix(angle)


def _make_rotation(order, *angles):
    rotation = [0] * 3

    assert len(angles) == 3

    for angle_index, axis in enumerate(order):
        axis_index = ord(axis) - ord('x')

        rotation[axis_index] = angles[angle_index]

    return rotation
