import larian.physics.model as model


def transform(scene, target_coordinate_system):
    if scene.coordinate_system == target_coordinate_system:
        return False

    cs_transform = _CoordinateSystemTransform(scene.coordinate_system, target_coordinate_system)

    for body in scene.bodies.values():
        _transform_body(body, cs_transform)

    for constraint in scene.constraints.values():
        _transform_constraint(constraint, cs_transform)

    scene.coordinate_system = target_coordinate_system

    return True


class _CoordinateSystemTransform(object):
    def __init__(self, source_coordinate_system, target_coordinate_system):
        source_urv_axis_indices, source_view_is_backward = _make_urv_map(source_coordinate_system)
        target_urv_axis_indices, target_view_is_backward = _make_urv_map(target_coordinate_system)

        self.__source_axis_to_target = [-1] * 3

        self.__target_view_axis = target_urv_axis_indices[2]
        self.__view_multiplier = +1.0 if source_view_is_backward == target_view_is_backward else -1.0

        for source_axis_index, target_axis_index in zip(source_urv_axis_indices, target_urv_axis_indices):
            self.__source_axis_to_target[source_axis_index] = target_axis_index

    def apply_to_vector(self, vector, view_flip=True):
        result = model.Vector3()

        for i in (0, 1, 2):
            result[self.__source_axis_to_target[i]] = vector[i]

        if view_flip:
            result[self.__target_view_axis] *= self.__view_multiplier

        return result

    def apply_to_basis(self, basis):
        result = model.Transform.Basis()

        for i in (0, 1, 2):
            row = basis[i]
            row = self.apply_to_vector(row)
            result[self.__source_axis_to_target[i]] = row

        for i in (0, 1, 2):
            result[self.__target_view_axis][i] *= self.__view_multiplier

        return result

    def apply_to_transform(self, transform):
        result = model.Transform()

        result.translation = self.apply_to_vector(transform.translation)
        result.basis = self.apply_to_basis(transform.basis)

        return result

    def apply_to_limits(self, limits):
        result = model.Constraint.Limits()

        result.upper = self.apply_to_vector(limits.upper, view_flip=False)
        result.lower = self.apply_to_vector(limits.lower, view_flip=False)

        if result.lower[self.__target_view_axis] < result.upper[self.__target_view_axis]:
            view_a = result.upper[self.__target_view_axis] * self.__view_multiplier
            view_b = result.lower[self.__target_view_axis] * self.__view_multiplier

            result.lower[self.__target_view_axis] = min(view_a, view_b)
            result.upper[self.__target_view_axis] = max(view_a, view_b)

        return result

    def apply_to_axis(self, axis):
        return chr(ord('x') + self.__source_axis_to_target[ord(axis) - ord('x')])


def _make_urv_map(coordinate_system):
    up = None
    right = None
    view = None
    view_is_backward = False

    for axis_index, direction in enumerate(coordinate_system):
        if direction == model.EAxisDirection.UP:
            up = axis_index
        elif direction == model.EAxisDirection.RIGHT:
            right = axis_index
        elif direction == model.EAxisDirection.FORWARD:
            assert view is None

            view = axis_index
        else:
            assert direction == model.EAxisDirection.BACKWARD
            assert view is None

            view = axis_index
            view_is_backward = True

    return (up, right, view), view_is_backward


def _transform_body(body, cs_transform):
    shape_transform_function_name = '_transform_shape_' + body.shape.type
    shape_transform_function = globals().get(shape_transform_function_name)

    if shape_transform_function is not None:
        shape_transform_function(body.shape, cs_transform)

        body.shape.transform = cs_transform.apply_to_transform(body.shape.transform)

    body.transform = cs_transform.apply_to_transform(body.transform)


def _transform_constraint(constraint, cs_transform):
    constraint.transform = cs_transform.apply_to_transform(constraint.transform)
    constraint.angular_limits = cs_transform.apply_to_limits(constraint.angular_limits)
    constraint.linear_limits = cs_transform.apply_to_limits(constraint.linear_limits)


def _transform_shape_box(shape, cs_transform):
    shape.extents = cs_transform.apply_to_vector(shape.extents, view_flip=False)


def _transform_shape_capsule(shape, cs_transform):
    shape.axis = cs_transform.apply_to_axis(shape.axis)


def _transform_shape_cylinder(shape, cs_transform):
    shape.axis = cs_transform.apply_to_axis(shape.axis)


def _transform_shape_hull(shape, cs_transform):
    for i, point in enumerate(shape.points):
        shape.points[i] = cs_transform.apply_to_vector(point)


def _transform_shape_mesh(shape, cs_transform):
    for i, vertex in enumerate(shape.vertices):
        shape.vertices[i] = cs_transform.apply_to_vector(vertex)

    indices = shape.indices

    for i in range(0, len(indices), 3):
        indices[i], indices[i + 1], indices[i + 2] = indices[i + 2], indices[i + 1], indices[i]
