from larian import utils


ID = "http://dev.larian.com/schemas/physics-01.json"


class EAxisDirection(utils.Enum):
    UP = 'up'
    RIGHT = 'right'
    BACKWARD = 'backward'
    FORWARD = 'forward'


class EBodyType(utils.Enum):
    DYNAMIC = 'dynamic'
    KINEMATIC = 'kinematic'
    STATIC = 'static'


class EAxis(utils.Enum):
    X = 'x'
    Y = 'y'
    Z = 'z'


class EShapeType(utils.Enum):
    BOX = 'box'
    SPHERE = 'sphere'

    CYLINDER = 'cylinder'
    CAPSULE = 'capsule'

    HULL = 'hull'
    MESH = 'mesh'


class Vector3(utils.ArrayStruct):
    FIELDS = [
        utils.Field('x', float),
        utils.Field('y', float),
        utils.Field('z', float)
    ]


class Transform(utils.Struct):
    class Basis(utils.ArrayStruct):
        FIELDS = [
            utils.Field('x', Vector3),
            utils.Field('y', Vector3),
            utils.Field('z', Vector3)
        ]

    FIELDS = [
        utils.Field('basis', Basis),
        utils.Field('translation', Vector3)
    ]


_COMMON_SHAPE_FIELDS = [
    utils.Field('type', EShapeType),
    utils.Field('transform', Transform, default=Transform())
]


class ShapeBox(utils.Struct):
    FIELDS = _COMMON_SHAPE_FIELDS + [
        utils.Field('extents', Vector3)
    ]


class ShapeSphere(utils.Struct):
    FIELDS = _COMMON_SHAPE_FIELDS + [
        utils.Field('radius', float)
    ]


class ShapeCylinder(utils.Struct):
    FIELDS = _COMMON_SHAPE_FIELDS + [
        utils.Field('axis', EAxis, default=EAxis.Y),

        utils.Field('radius', float),
        utils.Field('length', float)
    ]


class ShapeCapsule(utils.Struct):
    FIELDS = ShapeCylinder.FIELDS


class ShapeHull(utils.Struct):
    FIELDS = _COMMON_SHAPE_FIELDS + [
        utils.Field('points', utils.ListType(Vector3))
    ]


class ShapeMesh(utils.Struct):
    FIELDS = _COMMON_SHAPE_FIELDS + [
        utils.Field('indices', utils.ListType(int)),
        utils.Field('vertices', utils.ListType(Vector3))
    ]


class CoordinateSystem(utils.ArrayStruct):
    FIELDS = [
        utils.Field('x', EAxisDirection, init=EAxisDirection.RIGHT),
        utils.Field('y', EAxisDirection, init=EAxisDirection.UP),
        utils.Field('z', EAxisDirection, init=EAxisDirection.BACKWARD)
    ]


class RigidBody(utils.Struct):
    FIELDS = [
        utils.Field('type', EBodyType),
        utils.Field('mass', float),
        utils.Field('friction', float),
        utils.Field('restitution', float),
        utils.Field('transform', Transform),
        utils.Field('shape', utils.ChoiceType('type', {
            EShapeType.BOX: ShapeBox,
            EShapeType.SPHERE: ShapeSphere,
            EShapeType.CYLINDER: ShapeCylinder,
            EShapeType.CAPSULE: ShapeCapsule,
            EShapeType.HULL: ShapeHull,
            EShapeType.MESH: ShapeMesh
        }))
    ]


class Constraint(utils.Struct):
    class Limits(utils.Struct):
        FIELDS = [
            utils.Field('lower', Vector3),
            utils.Field('upper', Vector3)
        ]

    FIELDS = [
        utils.Field('body_a_name', str),
        utils.Field('body_b_name', str),

        utils.Field('transform', Transform),

        utils.Field('angular_limits', Limits),
        utils.Field('linear_limits', Limits)
    ]


class Scene(utils.Struct):
    FIELDS = [
        utils.Field('coordinate_system', CoordinateSystem, default=CoordinateSystem()),
        utils.Field('bodies', utils.DictType(RigidBody)),
        utils.Field('constraints', utils.DictType(Constraint))
    ]
