import copy
import logging

from larian.physics import model
from larian import utils


def load(json):
    filtered_json = {k: v for k, v in json.items() if not k.startswith('_') and k != '$schema'}

    context = _Context()

    scene = _load_item(context, model.Scene, filtered_json)

    if _report_errors(context):
        return None

    return scene


def save(scene):
    context = _Context()

    json = _save_item(context, scene)

    if _report_errors(context):
        return None

    json['$schema'] = model.ID
    return json


class _Context(object):
    def __init__(self, path=['#'], errors=[]):
        self.__path = path
        self.errors = errors

    def validate(self, condition, message, *args, **kwargs):
        if not condition:
            self.error(message, *args, **kwargs)

        return condition

    def error(self, message, *args, **kwargs):
        message = message.format(*args, **kwargs)
        path = '/'.join(self.__path)

        self.errors.append("{}: {}".format(path, message))

    def __add__(self, component):
        if not isinstance(component, str):
            component = str(component)

        return _Context(self.__path + [component], self.errors)


def _load_item(context, item_type, data):
    r = context.validate(issubclass(item_type, utils.Struct),
                         "Invalid spec: should be either array or object, but got {}", item_type.__name__)

    if not r:
        return None

    item = item_type()

    if item.IS_ARRAY:
        r = context.validate(len(data) == len(item),
                             "Invalid number of items: {} expected, got {}", len(item), len(data))
    else:
        unknown_keys = data.keys() - item.keys()

        r = context.validate(len(unknown_keys) == 0,
                             "Schema mismatch for {}. Unknown data keys: {}", item_type.__name__, unknown_keys)

    if not r:
        return None

    for i, field in enumerate(item.FIELDS):
        if item.IS_ARRAY:
            field_data = data[i]
            field_context = context + i
        else:
            field_data = data.get(field.name)
            field_context = context + field.name

        if field_data is None:
            if field_context.validate(field.default is not None, "a required field missing"):
                item[i] = copy.deepcopy(field.default)
        else:
            item[i] = _load_item_field(field_context, field.type, field_data)

    return item


def _load_item_field(context, type_, data):
    if not _validate_type(context, type_, data):
        return None

    if isinstance(type_, utils.ListType):
        return [_load_item_field(context + i, type_.item_type, x) for i, x in enumerate(data)]
    elif isinstance(type_, utils.DictType):
        return {k: _load_item_field(context + k, type_.value_type, x) for k, x in data.items()}
    elif isinstance(type_, utils.ChoiceType):
        selected_type = _select_choice_type(context, type_, data)

        if selected_type is None:
            return None

        return _load_item_field(context, selected_type, data)
    elif issubclass(type_, utils.Enum):
        r = context.validate(data in type_.values, "Unexpected enum value: got {}, valid choices: {}",
                             data, ", ".join(type_.values))

        if not r:
            return None

        return data
    elif issubclass(type_, utils.Struct):
        return _load_item(context, type_, data)
    else:
        return type_(data)


def _get_data_type(field, field_data):
    if field.type is None:
        assert field.choice_id_field_name is not None
        id_ = field_data[field.choice_id_field_name]
        return field.choice_id_to_type_map[id_]

    assert field.type is not None

    return field.type


def _save_item(context, item):
    r = context.validate(isinstance(item, utils.Struct), "Complex object expected, but got {}", type(item).__name__)

    if not r:
        return False

    data = [] if item.IS_ARRAY else {}

    for i, field in enumerate(item.FIELDS):
        field_context = context + field.name
        field_value = item[i]

        if not field_context.validate(field_value is not None, "value is None"):
            continue

        if field.default is not None and field_value == field.default:
            continue

        field_data = _save_item_field(field_context, field.type, field_value)

        if field_data is None:
            continue

        if item.IS_ARRAY:
            data.append(field_data)
        else:
            data[field.name] = field_data

    return data


def _save_item_field(context, type_, value):
    if not _validate_type(context, type_, value, is_save=True):
        return None

    if isinstance(type_, utils.ListType):
        return [_save_item_field(context + i, type_.item_type, x) for i, x in enumerate(value)]
    elif isinstance(type_, utils.DictType):
        return {k: _save_item_field(context + k, type_.value_type, v) for k, v in value.items()}
    elif isinstance(type_, utils.ChoiceType):
        data = _save_item(context, value)

        if not _set_choice_type(context, data, type_, type(value)):
            return None

        return data
    elif issubclass(type_, utils.Enum):
        r = context.validate(value in type_.values, "Unknown enum value {} (expected one of: {})", value, type_.values)

        if not r:
            return None

        return value
    elif issubclass(type_, utils.Struct):
        return _save_item(context, value)
    else:
        return value


def _select_choice_type(context, type_, data):
    try:
        return type_.select(data)
    except utils.ChoiceType.NoIdError as e:
        context.error("Choice id field '{}' not found", str(e))
    except utils.ChoiceType.UnknownIdError as e:
        context.error("Choice id value '{}' is unknown", str(e))

    return None


def _set_choice_type(context, data, type_, value_type):
    try:
        type_.set_type(data, value_type)
        return True
    except utils.ChoiceType.UnknownTypeError as e:
        context.error("Unknown choice type {}".format(e))
        return False


def _report_errors(context):
    for error in context.errors:
        logging.error(error)

    return len(context.errors) > 0


def _validate_type(context, type_, value, is_save=False):
    if isinstance(type_, utils.ChoiceType):
        expected_type = type_.select(value)
        return _validate_type(context, expected_type, value, is_save)

    expected_type = None

    if isinstance(type_, utils.DictType):
        expected_type = dict
    elif isinstance(type_, utils.ListType):
        expected_type = list
    elif not is_save and issubclass(type_, utils.Struct):
        expected_type = list if type_.IS_ARRAY else dict
    elif issubclass(type_, utils.Enum):
        expected_type = str
    else:
        expected_type = type_

    return context.validate(isinstance(value, expected_type),
                            "value is expected to be of {} type, but got {} ({})",
                            expected_type.__name__, type(value).__name__, value)
