import os
import logging
import json
import jsonschema
import sys

from larian.physics import model


SCHEMA_NAME = 'physics-01.json'

if getattr(sys, 'frozen', False):
    SELFDIR = os.path.dirname(sys.executable)
    SCHEMA_PATH = os.path.join(SELFDIR, "schemas", SCHEMA_NAME)
else:
    SELFDIR = os.path.dirname(__file__)
    SCHEMA_PATH = os.path.join(SELFDIR, '../../schemas', SCHEMA_NAME)

    # For tests. TODO: make tests run on installed dir, after installation. That will be ultimate check
    if not os.path.exists(SCHEMA_PATH):
        SCHEMA_PATH = os.path.join(SELFDIR, '../../../schemas', SCHEMA_NAME)


with open(SCHEMA_PATH, 'r') as f:
    SCHEMA = json.load(f)


jsonschema.Draft4Validator.check_schema(SCHEMA)


def validate_json(json):
    validator = jsonschema.Draft4Validator(SCHEMA)

    success = True

    for error in validator.iter_errors(json):
        success = False

        if 'shape' in error.path and error.validator == 'oneOf' and 'type' in error.instance:
            type_ = error.instance['type']

            try:
                id_ = error.schema['oneOf'].index({'$ref': '#/definitions/Shape' + type_.title()})
            except ValueError:
                continue

            for suberror in error.context:
                if suberror.schema_path[0] == id_:
                    _report_schema_error(suberror.absolute_path, suberror.message)
        else:
            _report_schema_error(error.absolute_path, error.message)

    return success


def validate_scene(scene):
    r = True

    for name, body in scene.bodies.items():
        if body.type == model.EBodyType.STATIC:
            r &= _validate(body.mass == 0.0, "Rigid body {} is static, but has non-zero mass".format(name))
        elif body.type == model.EBodyType.DYNAMIC:
            r &= _validate(body.mass != 0.0, "Rigid body {} is dynamic, but has zero mass".format(name))

        r &= _validate(body.type == model.EBodyType.STATIC or body.shape.type != model.EShapeType.MESH,
                       "Rigid body {} has mesh collision shape, but it is only allowed for static objects (it's {})"
                       .format(name, body.type))

    for name, constraint in scene.constraints.items():
        body_a_name = constraint.body_a_name
        body_b_name = constraint.body_b_name

        r &= _validate(body_a_name in scene.bodies, "Constraint {} refers to unknown body {}".format(name, body_a_name))
        r &= _validate(body_b_name in scene.bodies, "Constraint {} refers to unknown body {}".format(name, body_b_name))

    return r


def _report_schema_error(path, message):
    logging.error("JSON validation error {}: {}".format('/' + '/'.join([str(x) for x in path]), message))


def _validate(condition, message):
    if not condition:
        logging.error(message)

    return condition
