import copy
import sys


if sys.version_info >= (3, 0):
    from .utils_py3 import add_metaclass as __add_metaclass  # noqa
else:
    from .utils_py2 import add_metaclass as __add_metaclass  # noqa


class __EnumMetaclass(type):
    @property
    def keys(cls):
        return set([x for x in dir(cls) if not x.startswith('_')])

    @property
    def values(cls):
        return set([getattr(cls, x) for x in cls.keys])


Enum = __add_metaclass(__EnumMetaclass)


class Struct(object):
    IS_ARRAY = False

    def __init__(self, *args, **kwargs):
        fields = object.__getattribute__(self, 'FIELDS')
        values = [None] * len(fields)
        name_map = {}

        object.__setattr__(self, '_Struct__values', values)
        object.__setattr__(self, '_Struct__name_to_index', name_map)

        for i, field in enumerate(fields):
            try:
                value = args[i]
            except IndexError:
                value = kwargs.get(field.name)

            if value is None:
                if field.default is not None:
                    value = copy.deepcopy(field.default)
                elif field.init is not None:
                    value = copy.deepcopy(field.init)
                else:
                    value = field.type()

            values[i] = value
            name_map[field.name] = i

    def items(self):
        for i, field in enumerate(self.FIELDS):
            yield field.name, self.__values[i]

    def values(self):
        return self.__values

    def keys(self):
        for field in self.FIELDS:
            yield field.name

    def __getattribute__(self, name):
        try:
            return object.__getattribute__(self, name)
        except AttributeError:
            index = object.__getattribute__(self, '_Struct__name_to_index').get(name)

            if index is None:
                raise

            return object.__getattribute__(self, '_Struct__values')[index]

    def __setattr__(self, name, value):
        index = self.__name_to_index.get(name)

        if index is None:
            object.__setattr__(self, name, value)
            return

        field = self.FIELDS[index]

        assert field.name == name

        self.__values[index] = value

    def __getitem__(self, key):
        index = self.__resolve_key(key)

        return self.__values[index]

    def __setitem__(self, key, value):
        index = self.__resolve_key(key)

        self.__values[index] = value

    def __len__(self):
        return len(self.__values)

    def __resolve_key(self, key):
        if isinstance(key, str):
            return self.__name_to_index[key]
        else:
            return key

    def __eq__(self, other):
        return self.__values == other.__values

    def __repr__(self):
        result = "[" if self.IS_ARRAY else "{"

        for i, field in enumerate(self.FIELDS):
            if not self.IS_ARRAY:
                result += field.name + ":"

            result += repr(self.__values[i])
            result += ","

        return result[:-1] + ("]" if self.IS_ARRAY else "}")


class ArrayStruct(Struct):
    IS_ARRAY = True


class Field(object):
    def __init__(self, name, type_, default=None, init=None):
        super(Field, self).__init__()

        self.name = name
        self.type = type_
        self.default = default
        self.init = init


class DictType(object):
    def __init__(self, value_type):
        super(DictType, self).__init__()

        self.value_type = value_type

    def __call__(self):
        return {}


class ListType(object):
    def __init__(self, item_type):
        super(ListType, self).__init__()

        self.item_type = item_type

    def __call__(self):
        return []


class ChoiceType(object):
    class NoIdError(KeyError):
        pass

    class UnknownIdError(KeyError):
        pass

    class UnknownTypeError(KeyError):
        pass

    def __init__(self, id_field_name, id_to_type_map):
        self.__id_field_name = id_field_name
        self.__id_to_type_map = id_to_type_map
        self.__type_to_id_map = {t: i for i, t in id_to_type_map.items()}

    def select(self, object_):
        try:
            id_ = object_[self.__id_field_name]
        except KeyError as e:
            raise ChoiceType.NoIdError(str(e))

        type_ = self.__id_to_type_map.get(id_)

        if type_ is None:
            raise ChoiceType.UnknownIdError(id_)

        return type_

    def set_type(self, data, value_type):
        id_ = self.__type_to_id_map.get(value_type)

        if id_ is None:
            raise ChoiceType.UnknownTypeError(str(value_type))

        data[self.__id_field_name] = id_

    def __call__(self):
        return None
