#!python3

import argparse
import io
import json
import logging
import os
import sys

import larian.physics.bullet
import larian.physics.model
import larian.physics.serializer
import larian.physics.coordinate_system
import larian.physics.validator

import larian.utils


class EFileType(larian.utils.Enum):
    JSON = "json"
    BULLET = "bullet"


def main():
    parser = argparse.ArgumentParser(
        description="Larian physics scene processor as a backend for exporters"
    )

    parser.add_argument("-i", "--input", metavar="<path>", default='-',
                        help="Use file at <path> as input (default: - for stdin)")

    parser.add_argument("--input-type", choices=EFileType.values, default=EFileType.JSON,
                        help="Treat input as <input-type> (default: json)")

    parser.add_argument("-o", "--output", metavar="<path>", default='-',
                        help="Use file at <path> as output (default: - for stdout)")

    parser.add_argument("--output-type", choices=EFileType.values, default=EFileType.JSON,
                        help="Write output as <output-type> (default: json)")

    parser.add_argument("--coordinate-system", default=None, metavar='<x-direction>_<y-direction>_<z-direction>',
                        type=_CoordinateSystemSpec,
                        help="Convert scene to coordinate system "
                             "expressed as 'up', 'right' and 'forward' or 'backward' "
                             "x, y, z logical direction map if needed")

    parser.add_argument('--no-validation', default=False, action='store_true',
                        help="Disable validation of converted data - useful to debug what's gone wrong")

    parser.add_argument("-v", "--verbose", action="count", default=0,
                        help="Be more verbose")

    parser.add_argument("-q", "--quiet", action="count", default=0,
                        help="Be less verbose")

    parser.add_argument('--log-pipe-format', action='store_true', default=False,
                        help="Output log messages in !log:<level_number>:<message> format, making it easier to parse")

    args = parser.parse_args()

    log_options = {
        "level": logging.WARN - (args.verbose - args.quiet)*10,
        "format": '[%(asctime)s] [%(levelname)s] %(message)s'
    }

    if args.log_pipe_format:
        log_options['format'] = '!log:%(levelno)d:%(message)s'
        log_options['stream'] = sys.stderr

    logging.basicConfig(**log_options)

    input_file = _open_input_file(args.input, args.input_type == EFileType.BULLET)
    output_file = _open_output_file(args.output, args.output_type == EFileType.BULLET)

    sys.path.append(os.path.join(os.path.dirname(__file__), '../../build/bin'))

    success = run(input_file,
                  args.input_type,
                  output_file,
                  args.output_type,
                  args.coordinate_system,
                  not args.no_validation)

    return 0 if success else 1


def run(input_file, input_type, output_file, output_type, target_coordinate_system, validation=True):
    scene = _read_scene(input_file, input_type, validation)

    if scene is None:
        return False

    if validation and not larian.physics.validator.validate_scene(scene):
        return False

    if target_coordinate_system is not None:
        larian.physics.coordinate_system.transform(scene, target_coordinate_system)

    assert not validation or larian.physics.validator.validate_scene(scene)

    return _write_scene(output_file, output_type, scene, validation)


def _CoordinateSystemSpec(string):
    parts = string.split('_')

    if len(parts) != 3:
        raise argparse.ArgumentTypeError('Coordinate system specification {} invalid: '
                                         'xdirection_ydirection_zdirection format expected'.format(string))

    if len(set(parts)) != len(parts):
        raise argparse.ArgumentTypeError('Coordinate system specification {} invalid: non-unique axis directions'
                                         .format(string))

    has_forward = False
    has_backward = False

    for part in parts:
        if part not in larian.physics.model.EAxisDirection.values:
            raise argparse.ArgumentTypeError('Coordinate system specification {} invalid: unknown direction {}'
                                             .format(part, string))

        has_forward |= larian.physics.model.EAxisDirection.FORWARD == part
        has_backward |= larian.physics.model.EAxisDirection.BACKWARD == part

    if has_backward and has_forward:
        raise argparse.ArgumentTypeError('Coordinate system specification invalid: '
                                         'forward and backward directions are mutually exclusive')

    return larian.physics.model.CoordinateSystem(*parts)


def _read_scene(input_file, input_type, validation=True):
    file_is_text = isinstance(input_file, io.TextIOBase)
    assert file_is_text == (input_type == EFileType.JSON)

    data = input_file.read()

    if input_type == EFileType.JSON:
        input_json = json.loads(data)
    else:
        input_json = larian.physics.bullet.convert_to_json(data)

    if validation and not larian.physics.validator.validate_json(input_json):
        return None

    return larian.physics.serializer.load(input_json)


def _write_scene(output_file, output_type, scene, validation=True):
    file_is_text = isinstance(output_file, io.TextIOBase)
    assert file_is_text == (output_type == EFileType.JSON)

    output_json = larian.physics.serializer.save(scene)

    assert not validation or larian.physics.validator.validate_json(output_json)

    if output_type == EFileType.JSON:
        data = json.dumps(output_json, indent=4, sort_keys=True)
    else:
        data = larian.physics.bullet.convert_from_json(output_json)

    try:
        output_file.write(data)
        output_file.flush()
    except IOError as e:
        logging.error("Can't write output file {} (format {}): {}".format(output_file.name, output_type, e))
        return False

    return True


def _open_input_file(path, binary):
    if path == '-':
        if binary:
            return sys.stdin.detach()
        else:
            return sys.stdin
    else:
        return open(path, "r" + ("b" if binary else ""))


def _open_output_file(path, binary):
    if path == '-':
        if binary:
            return sys.stdout.detach()
        else:
            return sys.stdout
    else:
        return open(path, "w" + ("b" if binary else ""))


if __name__ == '__main__':
    retval = main()
    sys.exit(retval)
